/* Demonstate controlling GPIO 2 as a xAP BSC endpoint
*/
#include "ets_sys.h"
#include "osapi.h"
#include "gpio.h"
#include "os_type.h"
#include "user_config.h"
#include "user_interface.h"
#include "espconn.h"
#include "mem.h"
#include "driver/uart.h"
#include "mqtt.h"
#include "xap.h"

static volatile os_timer_t client_timer; // also xAP heartbeat
static volatile os_timer_t info_timer; // xAP info timer

MQTT_Client mqttClient;

uint8_t state=0;
#define LED_GPIO 2
#define LED_GPIO_MUX PERIPHS_IO_MUX_GPIO2_U
#define LED_GPIO_FUNC FUNC_GPIO2

/// Decode allowable values for a "state=value" key pair.
int bscDecodeState(char *statestr) {
    int i;
    static char *value[] = {"on","off","true","false","yes","no","1","0","toggle"};
    static int stateL[] = {1,0,1,0,1,0,1,0,2};
    if (statestr == NULL) return -1;
    for(i=0; i < sizeof(value)/sizeof(value[0]); i++) {
            if(strcasecmp(statestr, value[i]) == 0)
                    return stateL[i];
    }
    return -1;
}

static void ICACHE_FLASH_ATTR heartbeat(void) {
    #ifdef PLATFORM_DEBUG
    ets_uart_printf("XAP heartbeart\r\n");
    #endif            	

    char buff[XAP_DATA_LEN];
    os_sprintf(buff, "xap-hbeat\n"
                "{\n"
                "v=12\n"
                "hop=1\n"
                "uid=%s\n"
                "class=xap-hbeat.alive\n"
                "source=%s\n"
                "interval=60\n"
                "}\n", XAP_UID, XAP_SOURCE);
	MQTT_Publish(&mqttClient, TOPIC_PUB, buff, strlen(buff), 0, 0);
}        

static void ICACHE_FLASH_ATTR infoEvent(char *class) {
    char buff[XAP_DATA_LEN];
    os_sprintf(buff, "xap-header\n"
                "{\n"
                "v=12\n"
                "hop=1\n"
                "uid=%s\n"
                "class=%s\n"
                "source=%s\n"
                "}\n"
				"output.state\n"
				"{\n"
				"state=%s\n"
				"}\n", XAP_UID, class, XAP_SOURCE, state ? "on" : "off");
	MQTT_Publish(&mqttClient, TOPIC_PUB, buff, strlen(buff), 0, 0);
}

void mqttConnectedCb(uint32_t *args)
{
	MQTT_Client* client = (MQTT_Client*)args;
	ets_uart_printf("MQTT: Connected\r\n");
	MQTT_Subscribe(client, TOPIC_SOURCE, 0);
}

void mqttDisconnectedCb(uint32_t *args)
{
	MQTT_Client* client = (MQTT_Client*)args;
	ets_uart_printf("MQTT: Disconnected\r\n");
}

void mqttPublishedCb(uint32_t *args)
{
	MQTT_Client* client = (MQTT_Client*)args;
	ets_uart_printf("MQTT: Published\r\n");
}

void mqttDataCb(uint32_t *args, const char* topic, uint32_t topic_len, const char *data, uint32_t data_len)
{
	MQTT_Client* client = (MQTT_Client*)args;

	xAPFrame frame;
	// *data is altered by this call.
	parseMsg(&frame, (unsigned char *)data, data_len);
	
	char *class = xapGetValue(&frame,"xap-header","class");
	if(class == NULL) return;	
	if(strcasecmp(class,"xAPBSC.query") == 0) {
	  infoEvent("xAPBSC.info");
	} else if(strcasecmp(class,"xAPBSC.cmd") == 0) {	
		int msgState = bscDecodeState(xapGetValue(&frame,"output.state.1","state"));
		if (msgState == -1) 
			return;
		if(msgState == 2) { // toggle
			state = state == 1 ? 0 : 1;
		} else {
			state = msgState;
		}
		GPIO_OUTPUT_SET(LED_GPIO, state ^ 1);
		infoEvent("xAPBSC.event");
	}
}

void xap_init() {
	MQTT_InitConnection(&mqttClient, MQTT_HOST, MQTT_PORT, 0);
	MQTT_InitClient(&mqttClient, XAP_SOURCE, "", "", 120, 1);

	//MQTT_InitLWT(&mqttClient, "/lwt", "offline", 0, 0);
	MQTT_OnConnected(&mqttClient, mqttConnectedCb);
	MQTT_OnDisconnected(&mqttClient, mqttDisconnectedCb);
	MQTT_OnPublished(&mqttClient, mqttPublishedCb);
	MQTT_OnData(&mqttClient, mqttDataCb);

    MQTT_Connect(&mqttClient);
	heartbeat();
	infoEvent("xAPBSC.info");
	
	os_timer_disarm(&client_timer);
	os_timer_setfn(&client_timer, (os_timer_func_t *)heartbeat, NULL);
	os_timer_arm(&client_timer, XAP_HEARTBEAT_INTERVAL * 1000, 1);

	os_timer_setfn(&info_timer, (os_timer_func_t *)infoEvent, "xAPBSC.info");
	os_timer_arm(&info_timer, 120000, 1);
}

static void ICACHE_FLASH_ATTR wait_for_ip(uint8 flag)
{
	LOCAL struct ip_info ipconfig;
    os_timer_disarm(&client_timer);

    switch(wifi_station_get_connect_status())
	{
    	case STATION_GOT_IP:
            wifi_get_ip_info(STATION_IF, &ipconfig);
            if( ipconfig.ip.addr != 0) {
    			#ifdef PLATFORM_DEBUG
            	ets_uart_printf("WiFi connected\r\n");
    			#endif            	
				xap_init();
            } else {
                os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
                os_timer_arm(&client_timer, 500, 0);
            }
			break;
		case STATION_WRONG_PASSWORD:
			#ifdef PLATFORM_DEBUG
			ets_uart_printf("WiFi connecting error, wrong password\r\n");
			#endif
            os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
            os_timer_arm(&client_timer, 1000, 0);
			break;
		case STATION_NO_AP_FOUND:
			#ifdef PLATFORM_DEBUG
			ets_uart_printf("WiFi connecting error, ap not found\r\n");
			#endif
            os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
            os_timer_arm(&client_timer, 500, 0);
			break;
		case STATION_CONNECT_FAIL:
			#ifdef PLATFORM_DEBUG
			ets_uart_printf("WiFi connecting fail\r\n");
			#endif
            os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
            os_timer_arm(&client_timer, 1000, 0);
			break;
		default:
			#ifdef PLATFORM_DEBUG
			ets_uart_printf("WiFi connecting...\r\n");
			#endif
            os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
            os_timer_arm(&client_timer, 1000, 0);
	}
}

static void ICACHE_FLASH_ATTR system_is_done(void){
	//Bringing up WLAN
	wifi_station_connect();
	wifi_station_dhcpc_start();
	if(wifi_station_get_auto_connect() == 0)
		wifi_station_set_auto_connect(1);
	//Wait for connection
	os_timer_disarm(&client_timer);
	os_timer_setfn(&client_timer, (os_timer_func_t *)wait_for_ip, NULL);
	os_timer_arm(&client_timer, 100, 0);
}

void ICACHE_FLASH_ATTR setup_wifi_st_mode(void)
{
	wifi_set_opmode((wifi_get_opmode()|STATION_MODE)&STATIONAP_MODE);
	struct station_config stconfig;
	wifi_station_disconnect();
	wifi_station_dhcpc_stop();
	if(wifi_station_get_config(&stconfig))
	{
		os_memset(stconfig.ssid, 0, sizeof(stconfig.ssid));
		os_memset(stconfig.password, 0, sizeof(stconfig.password));
		os_sprintf(stconfig.ssid, "%s", WIFI_SSID);
		os_sprintf(stconfig.password, "%s", WIFI_PASSWORD);
		if(!wifi_station_set_config(&stconfig))
		{
			#ifdef PLATFORM_DEBUG
			ets_uart_printf("ESP8266 not set station config!\r\n");
			#endif
		}
	}
	if(wifi_get_phy_mode() != PHY_MODE_11N)
		wifi_set_phy_mode(PHY_MODE_11N);
	#ifdef PLATFORM_DEBUG
	ets_uart_printf("ESP8266 in STA mode configured.\r\n");
	#endif
}

//Init function
void ICACHE_FLASH_ATTR user_init()
{	
	UARTInit();  // 115200,n,8,1
	os_delay_us(1000);
	
	setup_wifi_st_mode();
	
	// Configure pin as a GPIO
	PIN_FUNC_SELECT(LED_GPIO_MUX, LED_GPIO_FUNC);
	// state = 0 (off) - let pin float HIGH-1 (off)
	// state = 1 (on) - pull down to GND-0 (on)
	GPIO_OUTPUT_SET(LED_GPIO, state ^ 1); // invert for path to GND
	
    //Wait for system to be done.
    system_init_done_cb(&system_is_done);
    //system_os_task()
}
